import logging
from datetime import datetime, timezone

import sentry_sdk
from asyncpg import Record

from udata_hydra import config, context
from udata_hydra.db.check import Check

log = logging.getLogger("udata-hydra")


class ExceptionWithSentryDetails(Exception):
    """
    Custom exception which enriches Sentry with tags if available.
    """

    def __init__(
        self,
        message: str | None = None,
        step: str | None = None,
        resource_id: str | None = None,
        url: str | None = None,
        check_id: int | None = None,
        table_name: str | None = None,
        *args,
    ) -> None:
        self.step = step
        self.message = message
        if sentry_sdk.Hub.current.client:
            with sentry_sdk.new_scope() as scope:
                # scope.set_level("warning")
                scope.set_tags(
                    {
                        "resource_id": resource_id or "unknown",
                        "url": url or "unknown",
                        "check_id": check_id or "unknown",
                        "table_name": table_name or "unknown",
                    }
                )
                sentry_sdk.capture_exception(self)
        super().__init__(message, *args)


class ParseException(ExceptionWithSentryDetails):
    """Exception raised when an error occurs during parsing."""

    pass


class IOException(ExceptionWithSentryDetails):
    """Exception raised when an error occurs during IO operations."""

    pass


async def handle_parse_exception(e: ParseException, table_name: str, check: Record | None) -> None:
    """Specific ParseException handling. Store error if in a check context. Also cleanup :table_name: if needed."""
    db = await context.pool("csv")
    await db.execute(f'DROP TABLE IF EXISTS "{table_name}"')
    if check:
        if config.SENTRY_DSN:
            with sentry_sdk.new_scope():
                event_id = sentry_sdk.capture_exception(e)
        # e.__cause__ let us access the "inherited" error of ParseException (raise e from cause)
        # it's called explicit exception chaining and it's very cool, look it up (PEP 3134)!
        err = f"{e.step}:sentry:{event_id}" if config.SENTRY_DSN else f"{e.step}:{str(e.__cause__)}"
        await Check.update(
            check["id"],
            {"parsing_error": err, "parsing_finished_at": datetime.now(timezone.utc)},
        )
        log.error("Parsing error", exc_info=e)
    else:
        raise e



import json
import uuid

from aiohttp import web
from asyncpg import Record
from marshmallow import ValidationError

from udata_hydra.db.resource import Resource
from udata_hydra.schemas import ResourceDocumentSchema, ResourceSchema


async def get_resource(request: web.Request) -> web.Response:
    """Endpoint to get a resource from the DB
    Respond with a 200 status code and a JSON body with the resource data
    If resource is not found, respond with a 404 status code
    """

    try:
        resource_id = str(uuid.UUID(request.match_info["resource_id"]))
    except Exception as e:
        raise web.HTTPBadRequest(text=json.dumps({"error": str(e)}))

    resource: Record | None = await Resource.get(resource_id)
    if not resource:
        raise web.HTTPNotFound()

    return web.json_response(ResourceSchema().dump(dict(resource)))


async def get_resource_status(request: web.Request) -> web.Response:
    """Endpoint to get the current status of a resource from the DB.
    It is the same as get_resource but only returns the status of the resource, saving bandwith and processing time.
    Respond with a 200 status code and a JSON body with the resource status
    If resource is not found, respond with a 404 status code
    """
    try:
        resource_id = str(uuid.UUID(request.match_info["resource_id"]))
    except Exception as e:
        raise web.HTTPBadRequest(text=json.dumps({"error": str(e)}))

    resource: Record | None = await Resource.get(resource_id=resource_id, column_name="status")
    if not resource:
        raise web.HTTPNotFound()

    status: str | None = resource["status"]
    status_verbose: str = Resource.STATUSES[status]

    latest_check_endpoint = str(request.app.router["get-latest-check"].url_for())

    return web.json_response(
        {
            "resource_id": resource_id,
            "status": status,
            "status_verbose": status_verbose,
            "latest_check_url": f"{request.scheme}://{request.host}{latest_check_endpoint}?resource_id={resource_id}",
        }
    )


async def create_resource(request: web.Request) -> web.Response:
    """Endpoint to receive a resource creation event from a source
    Will create a new resource in the DB "catalog" table and mark it as priority for next crawling
    Respond with a 200 status code and a JSON body with a message key set to "created"
    If error, respond with a 400 status code
    """
    try:
        payload = await request.json()
        valid_payload: dict = ResourceSchema().load(payload)
    except Exception as err:
        raise web.HTTPBadRequest(text=json.dumps({"error": str(err)}))

    document: dict | None = valid_payload["document"]
    if not document:
        raise web.HTTPBadRequest(text="Missing document body")

    dataset_id = valid_payload["dataset_id"]
    resource_id = valid_payload["resource_id"]

    await Resource.insert(
        dataset_id=dataset_id,
        resource_id=resource_id,
        url=document["url"],
        priority=True,
    )

    return web.json_response(ResourceDocumentSchema().dump(dict(document)), status=201)


async def update_resource(request: web.Request) -> web.Response:
    """Endpoint to receive a resource update event from a source
    Will update an existing resource in the DB "catalog" table and mark it as priority for next crawling
    Respond with a 200 status code and a JSON body with a message key set to "updated"
    If error, respond with a 400 status code
    """

    try:
        payload = await request.json()
        valid_payload: dict = ResourceSchema().load(payload)
    except Exception as err:
        raise web.HTTPBadRequest(text=json.dumps({"error": str(err)}))

    document: dict | None = valid_payload["document"]
    if not document:
        raise web.HTTPBadRequest(text="Missing document body")

    dataset_id: str = valid_payload["dataset_id"]
    resource_id: str = valid_payload["resource_id"]

    await Resource.update_or_insert(dataset_id, resource_id, document["url"])

    return web.json_response(ResourceDocumentSchema().dump(document), status=200)


async def delete_resource(request: web.Request) -> web.Response:
    try:
        resource_id = str(uuid.UUID(request.match_info["resource_id"]))
    except Exception as e:
        raise web.HTTPBadRequest(text=json.dumps({"error": str(e)}))

    resource: Record | None = await Resource.get(resource_id=resource_id)
    if not resource:
        raise web.HTTPNotFound()

    # Mark resource as deleted in catalog table
    await Resource.delete(resource_id=resource_id)

    return web.HTTPNoContent()



import json
from datetime import datetime, timezone

from asyncpg import Record

from udata_hydra.crawl.calculate_next_check import calculate_next_check_date
from udata_hydra.crawl.helpers import get_content_type_from_header, is_valid_status
from udata_hydra.db.check import Check
from udata_hydra.db.resource import Resource
from udata_hydra.utils import queue, send


async def preprocess_check_data(dataset_id: str, check_data: dict) -> tuple[dict, dict | None]:
    """Preprocess a check data.

    Insert a new check in the DB with the provided check data before analysis, and update the resource status and priority.
    If the check data has changed compared to the previous check, it will also send the check to udata, and provide a first estimated next_check date to the new check.

    Args:
        check_data: the check data to insert in the DB.

    Returns:
        The updated check data as it has just been inserted in the DB.
        The previous check data, or None if it's the first check.
    """

    check_data["resource_id"] = str(check_data["resource_id"])

    last_check: dict | None = None
    last_check_record: Record | None = await Check.get_by_resource_id(check_data["resource_id"])
    if last_check_record:
        last_check = dict(last_check_record)

    has_changed: bool = await has_check_changed(check_data, last_check)
    if has_changed:
        queue.enqueue(
            send,
            dataset_id=dataset_id,
            resource_id=check_data["resource_id"],
            document={
                "check:available": is_valid_status(check_data.get("status")),
                "check:status": check_data.get("status"),
                "check:timeout": check_data["timeout"],
                "check:date": datetime.now(timezone.utc).isoformat(),
                "check:error": check_data.get("error"),
                "check:headers:content-type": await get_content_type_from_header(
                    check_data.get("headers", {})
                ),
                "check:headers:content-length": int(
                    check_data.get("headers", {}).get("content-length", 0)
                )
                or None,
            },
            _priority="high",
        )

    # Update resource following check:
    # Reset resource status so that it's not forbidden to be checked again.
    # Reset priority so that it's not prioritised anymore.
    await Resource.update(check_data["resource_id"], data={"status": None, "priority": False})

    check_data["next_check_at"] = calculate_next_check_date(has_changed, last_check, None)

    new_check: dict = await Check.insert(data=check_data, returning="*")

    return new_check, last_check


async def has_check_changed(check_data: dict, last_check_data: dict | None) -> bool:
    """Check if the check has changed compared to the last one"""

    is_first_check: bool = last_check_data is None
    status_has_changed = last_check_data and check_data.get("status") != last_check_data.get(
        "status"
    )
    status_no_longer_available = (
        last_check_data
        and is_valid_status(last_check_data.get("status"))
        and not is_valid_status(check_data.get("status"))
    )
    timeout_has_changed = last_check_data and check_data.get("timeout") != last_check_data.get(
        "timeout"
    )
    current_headers = check_data.get("headers", {})
    last_check_headers = (
        json.loads(last_check_data.get("headers"))
        if last_check_data and last_check_data.get("headers")
        else {}
    )
    content_has_changed = last_check_data and (
        current_headers.get("content-length") != last_check_headers.get("content-length")
        or current_headers.get("content-type") != last_check_headers.get("content-type")
    )

    # TODO: Instead of computing criterions here, store payload and compare with previous one.
    # It would make debugging easier.
    criterions = {
        "is_first_check": is_first_check,
        "status_has_changed": status_has_changed,
        "status_no_longer_available": status_no_longer_available,
        "timeout_has_changed": timeout_has_changed,
        "content_has_changed": content_has_changed,
    }

    return any(criterions.values())



from datetime import datetime, timezone

from asyncpg import Record

from udata_hydra import config, context
from udata_hydra.db.resource import Resource


async def select_rows_based_on_query(connection, q: str, *args) -> list[Record]:
    """
    A transaction wrapper around a select query q pass as param with *args.
    It first creates a temporary table based on this query,
    then fetches the selected rows and drops the temporary table.
    It finally returns the selected rows.
    """
    temporary_table = "check_urls"
    create_temp_select_table_query = (
        f"""CREATE TEMPORARY TABLE {temporary_table} AS {q} FOR UPDATE;"""
    )
    # Update resource status to CRAWLING_URL
    update_select_catalog_query = f"""
        UPDATE catalog SET status = 'CRAWLING_URL' WHERE resource_id in (select resource_id from {temporary_table});
    """
    async with connection.transaction():
        await connection.execute("BEGIN;")
        await connection.execute(create_temp_select_table_query, *args)
        await connection.execute(update_select_catalog_query)
        to_check: list[Record] = await connection.fetch(f"SELECT * FROM {temporary_table};")
        await connection.execute("COMMIT;")
    await connection.execute(f"DROP TABLE {temporary_table};")
    return to_check


async def select_batch_resources_to_check() -> list[Record]:
    """Select a batch of resources to check from the catalog
    - It first selects resources with priority=True
    - ...then resources without last check
    - and if the total number of selected resources is still less than the batch size, it will also add resources with outdated last check in the batch
    """
    context.monitor().set_status("Getting a batch from catalog...")

    pool = await context.pool()
    async with pool.acquire() as connection:
        excluded = Resource.get_excluded_clause()

        # first resources that are prioritised
        q = f"""
            SELECT * FROM (
                SELECT catalog.url, dataset_id, resource_id
                FROM catalog
                WHERE {excluded}
                AND priority = True
            ) s
            ORDER BY random() LIMIT {config.BATCH_SIZE}
        """
        to_check: list[Record] = await select_rows_based_on_query(connection, q)

        # then resources with no last check
        # (either because they have never been checked before, or because the last check has been deleted)
        if len(to_check) < config.BATCH_SIZE:
            q = f"""
                SELECT * FROM (
                    SELECT catalog.url, dataset_id, resource_id
                    FROM catalog
                    WHERE catalog.last_check IS NULL
                    AND {excluded}
                    AND priority = False
                ) s
                ORDER BY random() LIMIT {config.BATCH_SIZE}
            """
            to_check += await select_rows_based_on_query(connection, q)

        # if not enough for our batch size, handle resources with planned new check
        if len(to_check) < config.BATCH_SIZE:
            now = datetime.now(timezone.utc)
            limit = config.BATCH_SIZE - len(to_check)
            q = f"""
            SELECT * FROM (
                SELECT catalog.url, dataset_id, catalog.resource_id
                FROM catalog, checks
                WHERE catalog.last_check IS NOT NULL
                AND catalog.last_check = checks.id
                AND (checks.next_check_at <= $1 OR checks.next_check_at IS NULL)
                AND {excluded}
                AND catalog.priority = False
            ) s
            ORDER BY random() LIMIT {limit}
            """
            to_check += await select_rows_based_on_query(connection, q, now)

    return to_check
